# Test various formulas for "homz" using computations with random elements of the rational field.

def random_rational_list(n):
   s = set()
   num_bound = 10 + n
   den_bound = 10 + n
   while len(s) < n:
      q = QQ.random_element(num_bound, den_bound)
      s.add(q)
   return list(s)

def random_rational_list_special(n):
   s = set()
   num_bound = 15 + n
   den_bound = 15 + n
   while len(s) < n:
      q = QQ.random_element(num_bound, den_bound)
      if (q not in [0, 1, -1]) and (q not in s) and (1 / q not in s):
         s.add(q)
   return list(s)

def prod_difs(x, j):
   return prod(x[j] - x[k] for k in range(j) + range(j + 1, len(x)))

def hom_via_geom_progr(t, m):
   n = len(t)
   return sum(t[j] ** (n + m - 1) / prod_difs(t, j) for j in range(n))

def homz_via_lincomb_hom(z, m):
   n = len(z)
   m2 = floor(m / 2)
   hvalues = [hom_via_geom_progr(z, m - 2 * k) for k in range(m2 + 1)]
   return sum(((-1) ** k) * binomial(n + m - k - 1, k) * hvalues[k] for k in range(m2 + 1))

def test_homz_via_lincomb_hom_numer(n, mmax, verbos):
   if verbos >= 2:
      print('test_homz_via_lincomb_hom_numer, n = %d, mmax = %d' % (n, mmax))
   x = random_rational_list_special(n)
   y = list(x) + [1 / x[j] for j in range(n)]
   z = [x[j] + 1 / x[j] for j in range(n)]
   res0 = [hom_via_geom_progr(y, m) for m in range(mmax + 1)]
   hvalues = [hom_via_geom_progr(z, k) for k in range(mmax + 1)]
   res1 = [0] * (mmax + 1)
   for m in range(mmax + 1):
      m2 = floor(m / 2)
      res1[m] = sum(((-1) ** k) * binomial(n + m - k - 1, k) * hvalues[m - 2 * k] for k in range(m2 + 1))
   if verbos >= 3:
      print('res0 = ' + str(res0))
      print('res1 = ' + str(res1))
   return res0 == res1  

def big_test_homz_via_lincomb_hom_numer(nmax, mmax, nrep, verbos):
   if verbos >= 1:
      print('big_test_homz_via_lincomb_hom_numer, nmax = %d, mmax = %d, nrep = %d' % (nmax, mmax, nrep))
   tasks = [(n, r) for n in range(1, nmax + 1) for r in range(nrep)]
   return all(test_homz_via_lincomb_hom_numer(n, mmax, verbos) for (n, r) in tasks)

def homz_via_sum_of_quotients_chebU(z, m):
   n = len(z)
   return sum(chebyshev_U(m + n - 1, z[j] / 2) / prod_difs(z, j) for j in range(n))

def mymultisubsets(n, m):
   reps = sorted(list(range(n)) * m)
   s = list(Subsets(reps, m, submultiset = True))
   ms = []
   for a in s:
      b = [0] * n
      for j in range(m):
         b[a[j]] += 1
      ms.append(b)
   return ms

def one_chebU_product(x, degrees):
   n = len(x)
   return prod(chebyshev_U(degrees[j], x[j] / 2) for j in range(n))

def homz_via_sum_of_products_of_ChebU(z, m):
   n = len(z)
   ms = mymultisubsets(n, m)
   return sum(one_chebU_product(z, degrees) for degrees in ms)

def vandermonde(x):
   n = len(x)
   return prod(prod_difs(x[j : ], 0) for j in range(n - 1))

def uvandermonde(x, la):
   n = len(x)
   F = x[0].parent()
   a = matrix(F, n, n)
   for j in range(n):
      for k in range(n):
         d = la[j] + n - 1 - j
         a[j, k] = chebyshev_U(d, x[k] / 2)
   return det(a)

def uquotient(x, la):
   F = x[0].parent()
   return F(uvandermonde(x, la) / vandermonde(x))

def homz_via_uquotient(x, m):
   n = len(x)
   la = [m] + [0] * (n - 1)
   return uquotient(x, la)

def test_homz_various_formulas_numer(n, m, verbos):
   if verbos >= 2:
      print('test_homz_various_formulas_numer, n = %d, m = %d' % (n, m))
   z = random_rational_list(n)
   res0 = homz_via_sum_of_quotients_chebU(z, m)
   res1 = homz_via_lincomb_hom(z, m)
   res2 = homz_via_sum_of_products_of_ChebU(z, m)
   res3 = homz_via_uquotient(z, m)
   if verbos >= 3:
      print('res0 = ' + str(res0))
      print('res1 = ' + str(res1))
      print('res2 = ' + str(res2))
      print('res3 = ' + str(res3))
   return (res0 == res1) and (res0 == res2) and (res0 == res3)

def big_test_homz_various_formulas_numer(nmax, mmax, nrep, verbos):
   if verbos >= 1:
      print('big_test_homz_various_formulas_numer, nmax = %d, mmax = %d, nrep = %d' % (nmax, mmax, nrep))
   tasks = [(n, m, r) for n in range(1, nmax + 1) for m in range(mmax + 1) for r in range(nrep)]
   return all(test_homz_various_formulas_numer(n, m, verbos) for (n, m, r) in tasks)

def element_Catalan_triangle(n, k):
   numer = factorial(n + k) * (n - k + 1)
   denom = factorial(k) * factorial(n + 1)
   return numer / denom

def coef_hom_via_homz(n, m, k):
   return element_Catalan_triangle(n + m - 1 - k, k)

def test_hom_is_lincomb_homz_numer(n, mmax, verbos):
   if verbos >= 2:
      print('test_hom_is_lincomb_homz_numer, n = %d, mmax = %d' % (n, mmax))
   x = random_rational_list_special(n)
   xrecip = [1 / x[j] for j in range(n)]
   y = list(x) + xrecip
   z = [x[j] + 1 / x[j] for j in range(n)]
   # res0 = hom(x + 1 / x)
   # res1 = sum coefs * hom(x, 1 / x)
   res0 = [hom_via_geom_progr(z, m) for m in range(mmax + 1)]
   homvalues = [hom_via_geom_progr(y, m) for m in range(mmax + 1)]
   res1 = [0] * (mmax + 1)
   for m in range(mmax + 1):
      m2 = floor(m / 2)
      res1[m] = sum([coef_hom_via_homz(n, m, k) * homvalues[m - 2 * k] for k in range(m2 + 1)])
   if verbos >= 3:
      print('res0 = ' + str(res0))
      print('res1 = ' + str(res1))
   return res0 == res1

def big_test_hom_is_lincomb_homz_numer(nmax, mmax, nrep, verbos):
   if verbos >= 1:
      print('big_test_hom_is_lincomb_homz_numer, nmax = %d, mmax = %d, nrep = %d' % (nmax, mmax, nrep))
   tasks = [(n, r) for n in range(1, nmax + 1) for r in range(nrep)]
   return all(test_hom_is_lincomb_homz_numer(n, mmax, verbos) for (n, r) in tasks)

def test_homz_recur_numer(n, mmax, verbos):
   if verbos >= 1:
      print('test_homz_recur_numer, n = %d' % n)
   zext = random_rational_list_special(n + 1)
   z = zext[ : n]
   F = QQ
   homz_ext = [homz_via_sum_of_quotients_chebU(zext, m) for m in range(mmax + 2)]
   homz_short = [homz_via_sum_of_quotients_chebU(z, m) for m in range(mmax + 2)]
   lhs = homz_ext
   rhs0 = homz_short
   rhs1 = [F.zero()] + [zext[n] * homz_ext[j] for j in range(mmax + 1)]
   rhs2 = [F.zero(), F.zero()] + [- homz_ext[j] for j in range(mmax)]
   rhs = [rhs0[j] + rhs1[j] + rhs2[j] for j in range(mmax + 2)]
   if verbos >= 3:
      print('lhs = ' + str(lhs))
      print('rhs = ' + str(rhs))
   return lhs == rhs

def big_test_homz_recur_numer(nmax, mmax, nrep, verbos):
   if verbos >= 1:
      print('big_test_homz_recur_numer, nmax = %d, mmax = %d, nrep = %d' % (nmax, mmax, nrep))
   tasks = [(n, r) for n in range(1, nmax + 1) for r in range(nrep)]
   return all(test_homz_recur_numer(n, mmax, verbos) for (n, r) in tasks)

def test_difference_homz_numer(n, mmax, verbos):
   if verbos >= 1:
      print('test_difference_homz_numer, n = %d, mmax = %d' % (n, mmax))
   zext12 = random_rational_list(n + 2)
   zshort = zext12[ : n]
   zext1 = zshort + [zext12[n]]
   zext2 = zshort + [zext12[n + 1]]
   F = QQ
   lhs1 = [homz_via_sum_of_quotients_chebU(zext1, m + 1) for m in range(mmax + 1)]
   lhs2 = [homz_via_sum_of_quotients_chebU(zext2, m + 1) for m in range(mmax + 1)]
   lhs = [lhs1[m] - lhs2[m] for m in range(mmax + 1)]
   rhs0 = [homz_via_sum_of_quotients_chebU(zext12, m) for m in range(mmax + 1)]
   rhs = [(zext12[n] - zext12[n + 1]) * rhs0[m] for m in range(mmax + 1)]
   if verbos >= 3:
      print('lhs = ' + str(lhs))
      print('rhs = ' + str(rhs))
   return lhs == rhs

def big_test_difference_homz_numer(nmax, mmax, nrep, verbos):
   if verbos >= 1:
      print('big_test_difference_homz_numer, nmax = %d, mmax = %d, nrep = %d' % (nmax, mmax, nrep))
   tasks = [(n, r) for n in range(1, nmax + 1) for r in range(nrep)]
   return all(test_difference_homz_numer(n, mmax, verbos) for (n, r) in tasks)

import time

def all_tests_homz_numer(nmax, mmax, nrep):
   verbos = 2
   res = [0] * 5
   t = time.time()
   res[0] = big_test_homz_via_lincomb_hom_numer(nmax, mmax, nrep, verbos)
   res[1] = big_test_homz_various_formulas_numer(nmax, mmax, nrep, verbos)
   res[2] = big_test_hom_is_lincomb_homz_numer(nmax, mmax, nrep, verbos)
   res[3] = big_test_homz_recur_numer(nmax, mmax, nrep, verbos)
   res[4] = big_test_difference_homz_numer(nmax, mmax, nrep, verbos)
   total_result = all(res)
   t = time.time() - t
   print('time = %.2f seconds' % t)
   print('total result = ' + str(total_result))

# with nmax = 5, mmax = 15, nrep = 10, time < 60 sec
all_tests_homz_numer(5, 15, 10)

